#ifndef READ_PROCESSOR_H
#define READ_PROCESSOR_H

#include "header.h"
#include "macros.h"

class RawRead;

// This will handle all read related queries and manage the I/O.
class ReadProcessor {
 public:
  ReadProcessor(const string& path);
  ~ReadProcessor();

  // Records will be returned an a vector of quadruples: id, offset, lmar, rmar.
  const string GetID(int id);
  const string& GetLMar(int id, mar_types mt) const;
  const string& GetRMar(int id, mar_types mt) const;
  const string& GetMar(int id, mar_types mt) const;
  int GetMarLen(int id, mar_types mt) const;
  const IntIntMtTripleVec& GetRecords(int val);

  inline int ConsumedBy(int i) const { return used_[i]; }
  inline int Coverage() const { return cov_; }
  inline void Consume(int i, int seed_id)
  { prev_used_[i] = used_[i]; used_[i] = seed_id; ++num_used_[i]; }
  inline void Release(int i) { used_[i] = prev_used_[i]; --num_used_[i]; }
  inline void Reset(int i) { used_[i] = prev_used_[i] = -1; num_used_[i] = 0; }
  inline int NumUsed(int i) { return num_used_[i]; } 
  inline int GetMerFreq(int i) { return fmer_[i]; }
  int TotalPairs() const;
  //  inline int UnusedPairs() const { return total_ - used_.count(); }

  // Take 'str' and update 'fmer_'.
  void FreqCount(const string& str) const;
  void ProcessFiles();
  void DumpFile(int val) const;
  void DumpKMerFreq() const;

 private:
  
  // We can open only a limited number of files at a time. So we handle values
  // within a range at the same time.
  void ProcessWithinValues(int from, int to);

  // Process the paired reads. This will make the search of a contig in a seeded
  // read faster.
  void ProcessPair(const RawRead& rread);

  // Helper method for 'ProcessPair'.
  void ProcessRead(int id, const string& anchor, mar_types mt);

  // Helper method for 'ProcessRead'.
  void Add2IndexedFile(int val, int id, int offset, mar_types mt);

  // Load a file into memory. If not enough space, unload some before.
  void LoadFile(int val);

  // Total # of files.
  static const int fc_ = 1 << (HASH_LEN << 1);

  // # of records in each file.
  static int frecs_[fc_];

  // Frequency of each k-mer (k == HASH_LEN).
  static int fmer_[fc_];

  static int cov_, total_, from_val_, to_val_;

  static string fname_[fc_]; 

  // File handlers;
  static fstream* fp_[fc_];

  // Check whether a particular pair is used or not.
  //  static bitset<NUM_PAIR> used_;
  static int used_[NUM_PAIR], prev_used_[NUM_PAIR], num_used_[NUM_PAIR];

  // Stores records for in-memory files. We will store these records based on
  // the available memory. If the file 'x' is not in memory, then
  // inmem_files_[x] will be empty. A record is of the format: [id, offset,
  // mar_type].
  static IntIntMtTripleVec inmem_files_[fc_];

  // Hash table to retrieve the records from an in-memory file faster.
  //  hash_map<int, int> index_inmem_files_[fc_];

  // IntVec read_nos_;

  static vector<RawRead> raw_read_;

  typedef struct IntIntInt {
    int a, b, c;
  };
};

#endif  //  READ_PROCESSOR_H
