#include <cassert>
#include <fstream>
#include <iostream>

#include "assembler.h"
//#include "constants.h"
#include "contig.h"
#include "global.h"
#include "hash.h"
#include "overlap-graph.h"
#include "read-processor.h"
#include "utils-string.h"
#include "utils-file.h"
#include "visualize.h"

int Assembler::leaf_contig_count_ = -1;

Assembler::Assembler(const vector<SeededRead>& psreads, ReadProcessor* rp,
		     fstream* log, fstream* contigs_out, fstream* geo_out,
		     fstream* geom_out, fstream* bambus_out) {
  cout << "running with the following configurations:\n"
       << "overlap between reads\t\t\t" << Global::MIN_OVERLAP << endl
       << "substitution error allowed per read\t" << Global::SUB_2MAP_SEED
       << endl
       << "output thickness\t\t\t" << Global::DESIRED_THICKNESS << endl
       << "min contig len\t\t\t\t" << Global::DESIRED_CONTIG_LEN << endl
       << "reference len\t\t\t\t" << Global::REF_SEQ_LEN << endl
       << "read use\t\t\t\t" << Global::MAX_READ_USE << endl;

  read_processor_ = rp;
  psreads_ = &psreads;
  log_ = log;
  contigs_out_ = contigs_out;
  geo_out_ = geo_out;
  geom_out_ = geom_out;
  bambus_out_ = bambus_out;
#ifdef DEBUG
  seed_out_ = UtilsFile::OpenFile(PREFIX_OUT + "seed.out", "w");
#endif  // DEBUG
  seed_count_ = psreads.size();
#ifdef REUSE_CONTIG
  total_base_count_ = 0;
  target_base_count_ = Global::REF_SEQ_LEN * Global::DESIRED_THICKNESS;
#endif  // REUSE_CONTIG
}

Assembler::~Assembler() {
#ifdef DEBUG
  UtilsFile::CloseFile(seed_out_);
#endif  // DEBUG
}

// ISSUES: what if seeded reads overlap?
bool Assembler::Assemble() {
  SStrIntPairVec seeds;

  // Make clusters using the seeded reads.
  for (int i = 0; i < (int)sreads_.size(); ++i) {

    *log_ << "\n-----------------\ncreating clusters from seeded read: " 
	  << sreads_[i] << endl; 
    const SeededRead& sread = (*psreads_)[sreads_[i]];
    const ShortyString& seed = (*psreads_)[sreads_[i]].Str();
#ifdef DEBUG
    //    cout << "pairs left: " << read_processor_->UnusedPairs() << endl;
//    cout << "processing seed: " << seed << endl;
#endif  // DEBUG
    int seed_id = (*psreads_)[sreads_[i]].Id();
    *contigs_out_ << ">contig_asm_0" << endl;
    *contigs_out_ << seed << endl;
    *geom_out_ << ">contig_asm_0" << endl;
    *geom_out_ << seed_id << " " << seed << " -1 0 -1" << endl;
    seed.PrintScores(*geom_out_);
#ifndef SINGLE_STRAND  // Double strand mode.
    Cluster lcluster, rcluster;
    Contig lcontig(false, -1, -1, seed, seed_id, read_processor_);
    Contig rcontig(true, -1, -1, seed, seed_id, read_processor_);
    seeds.clear();
    seeds.push_back(SStrIntPair(seed, seed_id));
    MakeCluster(false, seeds, &lcluster, &rcluster, &lcontig, &rcontig);
    //#ifdef DEBUG
//    cout << "cluster size(s)(left-right): " << lcluster.Size() << ' '
//	 << rcluster.Size() << endl;
    //#endif  // DEBUG

    if (!lcontig.IsGoodSeed(seed_out_, rcontig)) {
      continue;
    }

    // Create contigs from 'cluster'.
    if (lcluster.Size()) {
      //      lcontig.PrintContigFormat(bambus_out_);
      MakeContigs(true, lcluster, sread);
    }
    if (rcluster.Size()) {
      //rcontig.PrintContigFormat(bambus_out_);
      MakeContigs(false, rcluster, sread);
    }
    lcontig.Add(rcontig);
//    lcontig.PrintContigFormat(bambus_out_);
    lcontig.PrintGeoFormat(geo_out_);
    //    return true;
#endif
  }
  ProcessNewSeeds();
  return true;
}

void Assembler::ProcessNewSeeds() {
  SStrIntPairVec seeds;
  while (true) {
    if (lseeds_.empty() && rseeds_.empty()) return;
    while (!lseeds_.empty()) {
      const SeededRead sread = lseeds_.top();
      lseeds_.pop();
      const ShortyString& seed = sread.Str();
#ifdef DEBUG
//      cout << "processing seed: " << seed << endl;
#endif  // DEBUG
      int seed_id = sread.Id();
#ifndef SINGLE_STRAND  // Double strand mode.
      Cluster lcluster, rcluster; 
      Contig lcontig(false, -1, -1, seed, seed_id, read_processor_);
      Contig rcontig(true, -1, -1, seed, seed_id, read_processor_);
      seeds.clear();
      seeds.push_back(SStrIntPair(seed, seed_id));
      MakeCluster(false, seeds, &lcluster, &rcluster, &lcontig, &rcontig);
      //#ifdef DEBUG
//      cout << "_cluster size(s)(left-right): " << lcluster.Size() << ' '
//	   << rcluster.Size() << endl;
      //#endif  // DEBUG
      if (!lcontig.IsGoodSeed(seed_out_, rcontig)) {
	continue;
      }
      if (lcluster.Size()) {
	//	lcontig.PrintContigFormat(bambus_out_);
	MakeContigs(true, lcluster, sread);
      }
      if (rcluster.Size()) {
	//rcontig.PrintContigFormat(bambus_out_);
	MakeContigs(false, rcluster, sread);
      }
      lcontig.Add(rcontig);
//      lcontig.PrintContigFormat(bambus_out_);
      lcontig.PrintGeoFormat(geo_out_);
#endif
    }
//    cout << "switching seed gear!!!\n";
    
    // Process the right seeds.
    while (!rseeds_.empty()) {
      //++count;
      const SeededRead sread = rseeds_.top();
      rseeds_.pop();
      const ShortyString& seed = sread.Str();
#ifdef DEBUG
//      cout << "processing seed: " << seed << endl;
#endif  // DEBUG
      int seed_id = sread.Id();
#ifndef SINGLE_STRAND  // Double strand mode.
      Cluster lcluster, rcluster;
      Contig lcontig(false, -1, -1, seed, seed_id, read_processor_);
      Contig rcontig(true, -1, -1, seed, seed_id, read_processor_);
      seeds.clear();
      seeds.push_back(SStrIntPair(seed, seed_id));
      MakeCluster(false, seeds, &lcluster, &rcluster, &lcontig, &rcontig);
//      cout << "_cluster size(s)(left-right): " << lcluster.Size() << ' '
//	   << rcluster.Size() << endl;
      if (!lcontig.IsGoodSeed(seed_out_, rcontig)) {
	continue;
      }
      if (lcluster.Size()) {
	//lcontig.PrintContigFormat(bambus_out_);
	MakeContigs(true, lcluster, sread);
      }
      if (rcluster.Size()) {
	//rcontig.PrintContigFormat(bambus_out_);
	MakeContigs(false, rcluster, sread);
      }
      lcontig.Add(rcontig);
//      lcontig.PrintContigFormat(bambus_out_);
      lcontig.PrintGeoFormat(geo_out_);
#endif
    }
  }
}

void Assembler::MakeCluster(bool validate, const SStrIntPairVec& seeds,
			    Cluster* lcluster, Cluster* rcluster,
			    Contig* lcontig, Contig* rcontig) {

  // At some point we were passing a set of seeds, lets turn it off as it makes
  // geography info processing difficult.
  assert(seeds.size() == 1);

  SStrIntPairVec::const_iterator its = seeds.begin();
  for (; its != seeds.end(); ++its) {
    const ShortyString& seed = its->first;
    int seed_id = its->second;
    //    for (int i = 0; i < seed.size(); ++i) {
    const IntIntMtTripleVec& records = read_processor_->GetRecords(seed.Str());
    IntIntMtTripleVec::const_iterator ir = records.begin();
    for (; ir != records.end(); ++ir) {
      int id = ir->first;
      int offset = ir->second;
      //if (offset >= seeds.size()) {
      //	cout << "wrong offset " << offset << " " << seed.size() << endl; 
      //	continue;
      // }
      //      assert(offset < (int)seeds.size());
      mar_types mt = ir->third;
#ifdef REUSE_CONTIG
      if (read_processor_->ConsumedBy(id) == seed_id) continue;
      if (mt == LMAR || mt == LMARRC) {
	lcontig->AddRead(id, offset, -1, mt);
      } else {
	rcontig->AddRead(id, offset, -1, mt);
      }
      read_processor_->Consume(id, seed_id);
      if (read_processor_->NumUsed(id) >= Global::MAX_READ_USE) continue;
#endif  // REUSE_CONTIG
#ifndef REUSE_CONTIG
      if (validate) {
	if (read_processor_->ConsumedBy(id) == seed_id) continue;
      } else {
	if (read_processor_->ConsumedBy(id) != -1) continue;
      }
#endif  // REUSE_CONTIG
      if (mt == LMAR || mt == LMARRC) {
	lcluster->AddIndex(id, 1, mt);
	//	lcontig->AddRead(id, offset, -1, mt);
      } else {
	rcluster->AddIndex(id, 1, mt);
	//rcontig->AddRead(id, offset, -1, mt);
      }
    }
  }
}
  
void Assembler::MakeContigs(bool lcluster, const Cluster& cluster,
			    const SeededRead& sread) {
  // Check whether we have already achieved the desired coverage. If so then
  // stop.
//  cout << total_base_count_ << ' ' <<  target_base_count_ << '\n';
  if (total_base_count_ > target_base_count_) return;
  int seed_id = sread.Id();
  int seed_gen = sread.Gen();
  int gen = lcluster ? seed_gen + 1 : seed_gen - 1;
  OverlapGraph g(lcluster, cluster, &sread, read_processor_, log_, bambus_out_);
  //  time_t tstart = clock();
  // g.MergeAndCut();
  vector<Contig> contigs;
  g.VisitGraph(&contigs);
  //clock_t tend = clock();
  //  cout << "cluster size: " << cluster.Size() << " time: "
  //     << (tend - tstart) / CLOCKS_PER_SEC << endl;
  //exit(1);
#ifndef REUSE_CONTIG
  g.ReUseReads();
#endif  // REUSE_CONTIG
  //  g.PrintContigs(&contigs);

  // See whether it generated too many contigs which is an indication of a
  // repeat or jump.
  int total_clen = 0;
  for (int i = 0; i < (int)contigs.size(); ++i) {
    total_clen += contigs[i].Str().size();
  }
  if (total_clen > sread.Size() + 6 * Global::INSERT_LEN_DEV) {
    return;
  }
  for (int i = 0; i < (int)contigs.size(); ++i) {
    Contig& contig = contigs[i];
    const ShortyString& str = contig.Str();
    total_base_count_ += str.size();
    int contig_id;
    if (str.size() >= SEED_LEN) {
      contig_id = seed_count_;
      *geom_out_ << ">contig_asm_" << seed_count_ << endl;
      *geom_out_ << seed_count_++ << " " << str << " " << seed_id << " " << gen
		 << " " << lcluster << endl;
      str.PrintScores(*geom_out_);
      if (lcluster) {
	lseeds_.push(SeededRead(str, seed_count_ - 1, gen));
      } else {
	rseeds_.push(SeededRead(str, seed_count_ - 1, gen));
      }
      contig.SetId(seed_count_ - 1);
    } else {
      contig_id = leaf_contig_count_;
      *geom_out_ << ">contig_asm_" << leaf_contig_count_ << endl;
      *geom_out_ << leaf_contig_count_ << str << " " << seed_id << " " << gen
		 << " " << lcluster << endl;
      str.PrintScores(*geom_out_);
      contig.SetId(leaf_contig_count_);
      --leaf_contig_count_;
//      contig.PrintContigFormat(bambus_out_);
      contig.PrintGeoFormat(geo_out_);
      //contig.PrintFastaFormat(contigs_out_);
    }
    //*contigs_out_ << ">contig_asm_" << contig_id << endl;
    //*contigs_out_ << str << endl;
    //contig.PrintContigFormat(bambus_out_);
    contig.PrintFastaFormat(contigs_out_);
    //contig.PrintGeoFormat(geo_out_);
  }
}
