// This is a standalone program to take geography information of contigs and
// use them to merge/order them.

#include <iostream>
#include <fstream>

#include "contig.h"
#include "global.h"
#include "macros.h"
#include "overlap-graph.h"
#include "read-processor.h"
#include "shorty-string.h"
#include "utils-file.h"
#include "utils-math.h"
#include "utils-string.h"

const string PREFIX = "out/";

struct contig {
  bool obsolete; // Means this contigs is already being colapsed with someone.
  bool dir;  // Which side of the seed.
  int id;  // An unique integer id.
  //  int left;  // Left bound of the gap with the seed.
  //int right;  // Right bound of the gap with the seed.
  int root;  // Seed id or parent.
  int bundle;  // Number of spanning pairs between this and 'root'.
  int dist;  // Distance between this and 'root'.
  int sdev;  // Standard deviation of 'dist'.
  int gen;
  int curr_pos; // When obsolete, tell with whom it is marged.
  int count;  // How many fat contigs in participated in.
  int last; // Last contig that made 'ext_str' starting at this node.
  int scratch;
  string str_id;  // A fasta like string id.
  ShortyString str;  // The contig itself.
  ShortyString ext_str;  // The extended contig starting at this node.

  // Mapping of constituting reads. Sorted by (offset, id, mt).
  IntIntMtTripleVec read_map;
  IntIntMtTripleVec ext_read_map;

  // Reads sorted by (id, offset, mt).
  IntIntMtTripleVec read_map_id;
  //  IntIntMtTripleVec ext_read_map;
  //  set<IntPair, IntPairEq> ext_ids;  // Contigs (with offsets) that made the extended contigs.
  //IntPairVec adj; // Adjacent contigs. Stores (id, offset) pair.
  inline int size() { return str.size(); }
};

vector<contig> contigs;
vector<contig> econtigs;
//vector<list<int> > lchildren;
//vector<list<int> > rchildren;

ReadProcessor* rp;

// Used by the graph
//====================
bool* visited;
IntVec* readin[MAR_TYPES];
IntTripleVec* adj;
int* lclen;
IntPairVec* lpath;
//====================

struct SortReadMap {
  bool operator () (const IntIntMtTriple& a, const IntIntMtTriple& b) {
    if (a.second == b.second) {
      if (a.first == b.first) {
	return a.third < b.third;
      } else {
	return a.first < b.first;
      }
    }
    return a.second < b.second;
  }
};

struct SortReadMapByID {
  bool operator () (const IntIntMtTriple& a, const IntIntMtTriple& b) {
    if (a.first == b.first) {
      if (a.second == b.second) {
	return a.third < b.third;
      }
      return a.second < b.second;
    }
    return a.first < b.first;
  }
};

void PrintAdj() {
  for (int i = 0; i < contigs.size(); ++i) {
    cout << i << ":" << adj[i].size() << endl;
    //     for (int j = 0; j < adj[i].size(); ++j) {
    //       cout << " " << adj[i][j].first << "(" << adj[i][j].second << ","
    // 	   << adj[i][j].third << ")";
    //     }
    //     cout << endl;
  }
  exit(0);
}

inline bool eq_read_map(const IntIntMtTriple& a, const IntIntMtTriple& b) {
  return a.first == b.first && a.second == b.second && a.third == b.third;
}

inline bool eq_adj_id(const IntTriple& a, const IntTriple& b) { 
  return a.first == b.first;
}

inline bool lt_adj_id(const IntTriple& a, const IntTriple& b) {
  return a.first < b.first;
}

inline bool lt_adj_bundle(const IntTriple& a, const IntTriple& b) {
  return a.third < b.third;
}

inline bool gt_adj(const IntPair& a, const IntPair& b) {
  return a.first > b.first;
}

inline bool gt_contig_lens(const IntTriple& a, const IntTriple& b) {
  if (a.first == b.first) return a.third > b.third;
  return a.first > b.first;
}

inline bool lt_pair(const IntPair& a, const IntPair& b) {
  if (a.first == b.first) {
    return a.second < b.second;
  }
  return a.first < b.first;
}

inline bool lt_pair1(const pair<int, bool>& a, const pair<int, bool>& b) {
  return a.first < b.first;
}

inline bool eq_pair(const IntPair& a, const IntPair& b) {
  return a.first == b.first && a.second == b.second;
}

void PrintReadMapping(const IntIntMtTripleVec& x) {
  for (int i = 0; i < x.size(); ++i) {
    //cout << x[i].first << " ";
    cout << "<" << x[i].first << "," << x[i].second << "," << (int)x[i].third << ">";
  }
  cout << endl << "============\n";
}

void FindCommonReads(const vector<contig>& c, int a, int b,
		     IntIntMtTripleVec* ar, IntIntMtTripleVec* br) {
  int pa = 0;
  int pb = 0;
  while (pa < c[a].read_map_id.size() && pb < c[b].read_map_id.size()) {
    const IntIntMtTriple& ra = c[a].read_map_id[pa];
    const IntIntMtTriple& rb = c[b].read_map_id[pb];
    if (ra.first == rb.first) {
      if (ra.third == rb.third) {
	ar->push_back(ra);
	br->push_back(rb);
      }
      ++pa;
      ++pb;
    } else if (ra.first < rb.first) {
      ++pa;
    } else {
      ++pb;
    }
  }
  sort(ar->begin(), ar->end(), SortReadMap());
  sort(br->begin(), br->end(), SortReadMap());
}

int Dist(const IntIntMtTripleVec& ar, const IntIntMtTripleVec& br, int* sd, int* bundle) {
  assert(ar.size() == br.size());
  IntVec x;
  long long td = 0;
  for (int i = 0; i < ar.size(); ++i) {
    td += ar[i].second - br[i].second - rp->GetMarLen(br[i].first, br[i].third);
    x.push_back(ar[i].second - br[i].second - rp->GetMarLen(br[i].first, br[i].third));
    //cout << d << endl;
  }
  td /= (long long)ar.size();
  td /= 2; // This is the cutoff to remove the outliers.
  long long d = 0;
  int count = 0;
  for (int i = 0; i < x.size(); ++i) {
    if (x[i] >= td) {
      ++count;
      d += x[i];
    }
  }
  if (count) {
    d /= count;
  } else {
    return INF;
  }
  //cout << d << endl;
  d += (long long)Global::INSERT_LEN;
  //cout << d << endl;
  *sd = Global::INSERT_LEN_DEV / (int)sqrt(count);
  //exit(0);
  *bundle = count;
  return (int)d;
}

void FindCommonSpanningReads(const vector<contig>& c, int a, int b,
			     IntIntMtTripleVec* ar, IntIntMtTripleVec* br) {
  int pa = 0;
  int pb = 0;
  while (pa < c[a].read_map_id.size() && pb < c[b].read_map_id.size()) {
    const IntIntMtTriple& ra = c[a].read_map_id[pa];
    const IntIntMtTriple& rb = c[b].read_map_id[pb];
    if (ra.first == rb.first) {
      if ((ra.third == LMAR || ra.third == LMARRC) &&
	  mar_type_pair[ra.third] == rb.third) {
	ar->push_back(ra);
	br->push_back(rb);
      }
      ++pa;
      ++pb;
    } else if (ra.first < rb.first) {
      ++pa;
    } else {
      ++pb;
    }
  }
  sort(ar->begin(), ar->end(), SortReadMap());
  sort(br->begin(), br->end(), SortReadMap());
}

void ChimeraTest(const IntIntMtTripleVec& reads, int id, IntVec* cut_points) {
  cout << "##" << id << endl; 
  cut_points->clear();
  if (lclen[id] < Global::INSERT_LEN - 3 * Global::INSERT_LEN_DEV) {
    cut_points->push_back(0);
    cut_points->push_back(lclen[id]);
    return;
  }
  //   IntIntMtTripleVec reads;
  //   for (int i = 0; i < lpath[id].size(); ++i) {
  //     int x = lpath[id][i].first;
  //     int y = lpath[id][i].second;
  //     for (int j = 0; j < contigs[x].read_map.size(); ++j) {
  //       reads.push_back(IntIntMtTriple(contigs[x].read_map[j].first,
  // 				     contigs[x].read_map[j].second + y,
  // 				     contigs[x].read_map[j].third));
  //     }
  //   }
  //   sort(reads.begin(), reads.end(), SortReadMap());
  //   IntIntMtTripleVec::iterator new_end = unique(reads.begin(), reads.end(),
  // 					       eq_read_map);
  //   reads.erase(new_end, reads.end());
  IntPairVec l, r;
  for (int i = 0; i < reads.size(); ++i) {
    const IntIntMtTriple& x = reads[i];
    // assert(x.second >= 0);
    if (x.third == LMAR || x.third == LMARRC) {
      l.push_back(IntPair(x.first, x.second));
    } else {
      r.push_back(IntPair(x.first, x.second));
    }
  }
  sort(l.begin(), l.end(), lt_pair);
  sort(r.begin(), r.end(), lt_pair);
  IntPairVec::iterator new_end_pos = unique(l.begin(), l.end(), eq_pair);
  l.erase(new_end_pos, l.end());
  new_end_pos = unique(r.begin(), r.end(), eq_pair);
  r.erase(new_end_pos, r.end());
  vector<pair<int, bool> > event_points;
  int pl = 0;
  int pr = 0;
  while (pl < l.size() && pr < r.size()) {
    if (l[pl].first == r[pr].first) {
      int left = min(l[pl].second, r[pr].second);
      int right = max(l[pl].second, r[pr].second);

      // CAUTION
      // Not a correct estimate if the read lengths are different. Easy to fix.
      if (right - left <= Global::INSERT_LEN + 3 * Global::INSERT_LEN_DEV &&
	  right - left >= Global::INSERT_LEN - 3 * Global::INSERT_LEN_DEV) {
	event_points.push_back(pair<int, bool>(left, true));
	event_points.push_back(pair<int, bool>(right, false));
      }
      ++pl;
      ++pr;
    } else if (l[pl].first < r[pr].first) {
      ++pl;
    } else {
      ++pr;
    }
  }
  sort(event_points.begin(), event_points.end(), lt_pair1);
  int depth = 0;
  IntVec depths;
  for (int i = 0; i < event_points.size(); ++i) {
    if (event_points[i].second) ++depth;
    else --depth;
    //    cout << "[" << event_points[i].first << "," << depth << "]";
    cout << event_points[i].first << " " << depth << endl;      
    //    if (depth <= 0) {
    //  return true;
    // }
    depths.push_back(depth);
  }
  if (depths.size() == 0) {
    cut_points->push_back(0);
    cut_points->push_back(lclen[id]);    
    return;
  }
  long double avg = 0;
  for (int i = 0; i < depths.size(); ++i) {
    avg += (long double)depths[i] / depths.size();
  }
  int cutline = (int)avg / 2;
  cout << "cut line: " << cutline << endl;
  bool up = true;
  for (int i = 0; i < depths.size(); ++i) {
    if (up) { // looking for the up point.
      if (depths[i] > cutline) {
	up = false;
	cut_points->push_back(event_points[i].first);
      }
    } else {  // looking for the down point.
      if (depths[i] <= cutline) {
	up = true;
	cut_points->push_back(event_points[i].first);
      }
    }
  }
}

bool OverlapTest(int x, int y, int* offseta, int* bundle) {
  IntIntMtTripleVec cr;
  ShortyString c;
  return UtilsString::OverlapFromReadMapping(contigs[x].str,
  					     contigs[y].str,
  					     contigs[x].read_map,
  					     contigs[y].read_map,
  					     rp, offseta, bundle, &c, &cr);
  IntIntMtTripleVec ar, br;
  FindCommonReads(contigs, x, y, &ar, &br);
  bool overlap = 
    UtilsString::OverlapFromCommonReadMapping(contigs[x].str,
					      contigs[y].str, ar, br, rp,
					      offseta, bundle, &c);
  if (!overlap) return false;
  // This bundle might be very weak. Lets check it.
  int start = -1, end = -1;
  for (int i = 0; i < contigs[x].read_map.size(); ++i) {
    if (contigs[x].read_map[i].second == *offseta && start == -1) {
      start = i;
    }
    if (start != -1 &&
	contigs[x].read_map[i].second +
	rp->GetMarLen(contigs[x].read_map[i].first,
		      contigs[x].read_map[i].third)
	> contigs[x].read_map[start].second +
	contigs[y].size()) {
      break;
    }
    end = i;
  }
  //  assert(start <= end);
  int suba = end - start + 1 - *bundle;
  int endb = -1;
  for (int i = 0; i < contigs[y].read_map.size(); ++i) {
    if (contigs[y].read_map[i].second + *offseta +
	rp->GetMarLen(contigs[y].read_map[i].first,
		      contigs[y].read_map[i].third)
	<= contigs[x].read_map[end].second +
	rp->GetMarLen(contigs[x].read_map[end].first,
		      contigs[x].read_map[end].third)) {
      endb = i;
    } else break;
  }
  int subb = endb + 1 - *bundle;
  *bundle -= (suba + subb);
  return *bundle > 0;
}

bool CommonReadOverlapTest(int x, int y, int* offseta, int* bundle) {
  //  IntIntMtTripleVec cr;
  ShortyString c;
  // return UtilsString::OverlapFromReadMapping(contigs[x].str,
  //    					     contigs[y].str,
  // 					     contigs[x].read_map,
  // 					     contigs[y].read_map,
  //					     rp, offseta, bundle, &c, &cr);
  IntIntMtTripleVec ar, br;
  FindCommonReads(contigs, x, y, &ar, &br);
  bool overlap = 
    UtilsString::OverlapFromCommonReadMapping(contigs[x].str,
					      contigs[y].str, ar, br, rp,
					      offseta, bundle, &c);
  if (!overlap) return false;
  // This bundle might be very weak. Lets check it.
  int start = -1, end = -1;
  for (int i = 0; i < contigs[x].read_map.size(); ++i) {
    if (contigs[x].read_map[i].second == *offseta && start == -1) {
      start = i;
    }
    if (start != -1 &&
	contigs[x].read_map[i].second +
	rp->GetMarLen(contigs[x].read_map[i].first,
		      contigs[x].read_map[i].third)
	> contigs[x].read_map[start].second +
	contigs[y].size()) {
      break;
    }
    end = i;
  }
  //  assert(start <= end);
  int suba = end - start + 1 - *bundle;
  int endb = -1;
  for (int i = 0; i < contigs[y].read_map.size(); ++i) {
    if (contigs[y].read_map[i].second + *offseta +
	rp->GetMarLen(contigs[y].read_map[i].first,
		      contigs[y].read_map[i].third)
	<= contigs[x].read_map[end].second +
	rp->GetMarLen(contigs[x].read_map[end].first,
		      contigs[x].read_map[end].third)) {
      endb = i;
    } else break;
  }
  int subb = endb + 1 - *bundle;
  *bundle -= (suba + subb);
  return *bundle > 0;
}

void BuildGraph() {
  for (int i = 0; i < MAR_TYPES; ++i) {
    readin[i] = new IntVec[rp->TotalPairs()];
  }
  visited = new bool[contigs.size()];
  adj = new IntTripleVec[contigs.size()];
  lpath = new IntPairVec[contigs.size()];
  lclen = new int[contigs.size()];
  for (int i = 0; i < contigs.size(); ++i) {
    visited[i] = false;
    lclen[i] = contigs[i].size();
    lpath[i].push_back(IntPair(i, 0));
    if (contigs[i].obsolete) continue;
    for (int j = 0; j < contigs[i].read_map.size(); ++j) {
      const IntIntMtTriple& x = contigs[i].read_map[j];
      if (x.second == 0) {
	readin[x.third][x.first].push_back(i);
      } else {
	break;
      }
    }
  }
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].obsolete) continue;
    int prev = -1;
    for (int j = 0; j < contigs[i].read_map.size(); ++j) {
      const IntIntMtTriple& x = contigs[i].read_map[j];
      for (int k = 0; k < readin[x.third][x.first].size(); ++k) {
	int id = readin[x.third][x.first][k];
	if (id == i) continue;
	if (id == prev) continue;
	int offseta, bundle;
	if (OverlapTest(i, id, &offseta, &bundle)) {
	  //assert(offseta >= 0);
	  adj[i].push_back(IntTriple(id, offseta, bundle));
	  prev = id;
	}
      }
    }
  }
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].obsolete) continue;
    sort(adj[i].begin(), adj[i].end(), lt_adj_id);
    IntTripleVec::iterator new_end_pos = unique(adj[i].begin(),
						adj[i].end(), eq_adj_id);
    adj[i].erase(new_end_pos, adj[i].end());
    sort(adj[i].begin(), adj[i].end(), lt_adj_bundle);
  }
  cout << "graph building finished\n";
  //PrintAdj();
}

void Traverse(int curr) {
  stack<IntPair> nodes;  // root + depth
  stack<IntPair> tnodes;  // elegth+ poffset
  nodes.push(IntPair(curr, 0));
  tnodes.push(IntPair(contigs[curr].size(), 0));
  IntPairVec path;
  while (nodes.size()) {
    int root = nodes.top().first;
    int depth = nodes.top().second;
    int elength = tnodes.top().first;
    int poffset = tnodes.top().second;
    nodes.pop();
    tnodes.pop();
    if (visited[root]) {  // Due to cycles, it was already visited.
      continue;
    }
    visited[root] = true;
    cout << "visited " << root << endl;
    if (path.size() >= depth + 1) {
      path[depth] = IntPair(root, poffset);
    } else {
      path.push_back(IntPair(root, poffset));
    }
    bool any  = false;
    for (int i = 0; i < adj[root].size(); ++i) {
      int j = adj[root][i].first;
      if (visited[j]) {
	int add = (contigs[root].size() - adj[root][i].second) > lclen[j] ? 0
	  : (lclen[j] - contigs[root].size() + adj[root][i].second);
	int newlen = elength - poffset < contigs[root].size() + add ?
	  (poffset + contigs[root].size() + add) : elength;
	int start = path[0].first;
	if (newlen >= lclen[start]) {
	  //contigs[j].obsolete = false;
	  lclen[start] = newlen;
	  IntPairVec tpath; // Need this to avoid a cycle.
	  //lpath_[start].clear();
	  for (int k = 0; k <= depth; ++k) {
	    tpath.push_back(path[k]);
	  }
	  for (int k = 0; k < lpath[j].size(); ++k) {
	    tpath.push_back(IntPair(lpath[j][k].first,
				    adj[root][i].second + lpath[j][k].second +
				    poffset));
	  }
	  lpath[start] = tpath;
	}
	continue;
      }
      any = true;
      int add = (contigs[root].size() - adj[root][i].second) > contigs[j].size()
	? 0 : (contigs[j].size() - contigs[root].size() + adj[root][i].second);
      int newlen = elength - poffset < contigs[root].size() + add ? 
	(poffset + contigs[root].size() + add) : elength;
      nodes.push(IntPair(adj[root][i].first, depth + 1));
      tnodes.push(IntPair(newlen, poffset + adj[root][i].second));
      //break;  // We only want one path.
    }
    if (!any) {
      int start = path[0].first;
      if (elength > lclen[start] ||
	  (elength == lclen[start] && depth >= lpath[start].size())) {
	lclen[start] = elength;
	lpath[start].clear();
	for (int i = 0; i <= depth; ++i) {
	  lpath[start].push_back(path[i]);
	  //	  if (i) contigs[path[i].first].obsolete = true;
	}
      } else {
	// Find the node where this path diverted. That might well be 
	// considered a sucpicious node making a chimera.
	for (int i = 1; i <= depth && i < lpath[start].size(); ++i) {
	  int id = path[i].first;
	  if (id != lpath[start][i].first) {
	    //int clen = elength - path[i].second;
	    int soffset = path[i].second;
	    int clen = contigs[id].size();
	    for (int j = i + 1; j <= depth; ++j) {
	      clen = max(clen, path[j].second - soffset +
			 contigs[path[j].first].size());
	    }
	    if (clen > lclen[id]) {
	      lclen[id] = clen;
	      lpath[id].clear();
	      for (; i <= depth; ++i) {
		lpath[id].push_back(IntPair(path[i].first,
					    path[i].second - soffset));
	      }
	    }
	    break;
	  }
	}
      }
    }
  }
}

void VisitGraph() {
  BuildGraph();
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].obsolete) continue;
    if (visited[i]) continue;
    Traverse(i);
    cout << "traversed " << i << endl;
  }

  //   // Free the graph related data structures.
  //   delete visited;
  //   for (int i = 0; i < MAR_TYPES; ++i) {
  //     delete readin[i];
  //   }
}

void PrintConfigInfo() {
  cout << "running with the following configurations:\n"
    //  << "second phase\t\t\t\t" << second_phase << endl
       << "overlap between contigs\t\t\t" << Global::MIN_OVERLAP << endl
       << "min score for overlap\t\t\t" << Global::MIN_SCORE << endl
       << "min contig len\t\t\t\t" << Global::DESIRED_CONTIG_LEN << endl;
}

void AnalyzeContig(int id) {
  cout << lpath[id][0].first;
  for (int i = 1; i < lpath[id].size(); ++i) {
    int x = lpath[id][i - 1].first;
    int y = lpath[id][i].first;
    int offseta, bundle;
    bool overlap = OverlapTest(x, y, &offseta, &bundle);
    if (!overlap) {
      //      cout << endl << x << "!o" << y << endl;
      //exit(1);
      cout << "(-)";
    } else {
      cout << "(" << bundle << ")";
    }
    cout << y;
  }
  cout << endl;
}

// void GetOverlappingContigs(bool dir, int root, IntVec* ids) {
//   const list<int>& target = dir ? rchildren[root] : lchildren[root];
//   const vector<list<int> >& other = dir ? lchildren : rchildren;
//   list<int>::const_iterator it = target.begin();
//   for (; it != target.end(); ++it) {
//     // Take the left (right) children of every right (left) child.
//     list<int>::const_iterator ito = other[*it].begin();
//     for (; ito != other[*it].end(); ++ito) {
//       if (contigs[root].curr_pos == contigs[*ito].curr_pos) continue;
//       ids->push_back(*ito);
//       GetOverlappingContigs(dir ? false : true, *ito, ids);
//     }
//   }
// }

// int total_candidate = 0;
// int success = 0;

// void EnrichContig(int cid) {
//   if (contigs[cid].obsolete) return;
//   IntVec ids;
//   GetOverlappingContigs(true, cid, &ids);  // Right
//   GetOverlappingContigs(false, cid, &ids);  // Left
//   cout << "possible overlaps for " << cid;
//   for (int i = 0; i < ids.size(); ++i) {
//     int ida = contigs[cid].curr_pos;
//     int idb = contigs[ids[i]].curr_pos;
//     if (ida == idb) continue;
//     ++total_candidate;
//     //   int score, offseta, overlap;
//     IntIntMtTripleVec cr;
//     ShortyString c;
//     bool overlap;
//     int offseta, bundle, first, second;
//     overlap = UtilsString::OverlapFromReadMapping(contigs[ida].ext_str,
// 						  contigs[idb].ext_str,
// 						  contigs[ida].read_map,
// 						  contigs[idb].read_map,
// 						  rp, &offseta, &bundle,
// 						  &c, &cr);
//     if (!overlap) {
//       overlap = UtilsString::OverlapFromReadMapping(contigs[idb].ext_str,
// 						    contigs[ida].ext_str,
// 						    contigs[idb].read_map,
// 						    contigs[ida].read_map,
// 						    rp, &offseta, &bundle,
// 						    &c, &cr);
//       if (overlap) {
// 	first = idb;
// 	second = ida;
//       }
//     } else {
//       first = ida;
//       second = idb;
//     }
//     cout << " " << ids[i];
//     if (overlap) {
//       // Merge the low lists.
//       lchildren[first].merge(lchildren[second]);
//       rchildren[first].merge(rchildren[second]);
//       contigs[first].ext_str = contigs[first].str = c;
//       contigs[first].read_map /*= contigs[first].ext_read_map*/ = cr;
//       contigs[first].obsolete = false;
//       contigs[second].curr_pos = first;
//       contigs[second].ext_str = contigs[second].str = c;
//       contigs[second].read_map /*= contigs[second].ext_read_map*/ = cr;
//       contigs[second].obsolete = true;
//       ++success;
//     } else {
//       cout << "(X)";
//     }
//   }
//   cout << endl;
// }

bool SortContigs(const contig& a, const contig& b) {
  if (a.root == b.root) {
    return a.dist < b.dist;
  }
  return a.root < b.root;
}

void UpdatePath(int i) {
  int curr = i;
  int dist = 0;
  
  // CAUTION: ingnoring issues related to a cycle.
  while (contigs[curr].root != -1) {
    contigs[curr].scratch = i;
    dist += contigs[curr].dist;
    curr = contigs[curr].root;
    if (contigs[curr].scratch == i) {
      contigs[i].root = -1;//contigs[i].id;
      contigs[i].dist = 0;
      return;
      // break;
    }
    contigs[i].root = curr;
    cout << " " << curr;
  }
  contigs[i].dist = dist;
}

double FractionCovered(int x, const IntIntMtTripleVec& r) {
  int prev = -1;
  int gap = 0;
  for (int i = 0; i < r.size(); ++i) {
    if (r[i].second <= prev + 1) {
    } else {
      gap += r[i].second - prev - 1;
    }
    prev = max(prev, r[i].second + rp->GetMarLen(r[i].first, r[i].third) - 1);
  }
  gap += contigs[x].size() - 1 - prev;
  cout << gap << endl;
  return 100.0 - 100.0 * gap / contigs[x].size();
}

int main(int argc, char* argv[]) {
  // Read the arguments.
  int c;
  while ((c = getopt(argc, argv, "o:l:")) != -1) {
    switch (c) {
    case 'o':
      Global::MIN_OVERLAP = atoi(optarg);
      Global::MIN_SCORE = 4 * Global::MIN_OVERLAP -
	min((int)(0.25 * 4 * Global::MIN_OVERLAP), 25);
      break;
    case 'l':
      Global::DESIRED_CONTIG_LEN = atoi(optarg);
      break;
    default:
      cout << "unknown option!!!\n";
      break;
    }
  }
  PrintConfigInfo();
  const string& read_file = argv[optind];
  rp = new ReadProcessor(read_file);
  const string& ext = argv[optind + 1];
  cout << "using output extension: " << ext << endl;
  int input_pair = atoi(argv[optind + 2]);
  cout << "total seed: " << input_pair << endl;
  int prev_total = 0;
  for (int i = 0; i < input_pair; ++i) {
    const string& geo_info = argv[optind + 2 * i + 3];
    const string& read_info = argv[optind + 2 * i + 4];
    cout << "using geom file: " << geo_info << endl;
    cout << "using geo file: " << read_info << endl;
    fstream* geo = UtilsFile::OpenFile(geo_info, "r");
    fstream* read_map = UtilsFile::OpenFile(read_info, "r");
    string id;
    int curr_total = 0;
    while (*geo >> id) {
      ++curr_total;
      int dir;
      contig temp;
      temp.str_id = id.substr(1, id.size() - 1);
      *geo >> temp.id >> temp.str >> temp.root >> temp.gen >> dir;
      temp.id += prev_total;
      //if (temp.id != contigs.size()) {
      //	cout << temp.str_id << " " << temp.id << " " << contigs.size() << endl;
      //}
      assert(temp.id == contigs.size());
      temp.str.ReadVotesFromFile(*geo);
      temp.dir = dir ? true : false;
      temp.obsolete = false;
      temp.curr_pos = temp.id;
      temp.ext_str = temp.str;
      //    temp.ext_ids.insert(IntPair(contigs.size(), 0));
      temp.count = 0;
      
      // Try to clean the contigs if deletions are allowed.
      if (PERCENTAGE_DEL_ALLOWED) {
	temp.str = temp.str.GetConfidentClone();
      }
      contigs.push_back(temp);
    }
    int iid;
    //cout << "len is: " << contigs[5122].size() << endl;
    while (*read_map >> iid) {
      iid += prev_total;
      int total_reads;
      *read_map >> total_reads;
      while (total_reads--) {
	int rid, roffset, mt;
	*read_map >> rid >> roffset >> mt;
	assert(roffset < contigs[iid].size());
	//	if (iid == 5122) cout << " " << roffset;
	//assert(roffset >= 0);
	contigs[iid].read_map.push_back(IntIntMtTriple(rid, roffset,
						       (mar_types)mt));
      }
      sort(contigs[iid].read_map.begin(), contigs[iid].read_map.end(),
	   SortReadMap());
      contigs[iid].read_map_id = contigs[iid].read_map;
      sort(contigs[iid].read_map_id.begin(), contigs[iid].read_map_id.end(),
	   SortReadMapByID());
    }
    prev_total += curr_total;
    UtilsFile::CloseFile(geo);
    UtilsFile::CloseFile(read_map);
  }
  cout << "total input contigs: " << contigs.size() << endl;
//   for (int i = 0; i < contigs.size(); ++i) {
//     ShortyString c;
//     UtilsString::MergeReads(contigs[i].read_map, contigs[i].size(), rp, &c);
//     cout << ">" << i << endl << c << endl;
//   }
//   exit(0);
  //int bundle, offseta;
  //bool overlap;
  //overlap = OverlapTest(3, 113, &offseta, &bundle);
  //cout << overlap << endl;
  //exit(0);
  //PrintReadMapping(contigs[3].read_map);
  //PrintReadMapping(contigs[113].read_map);
  //exit(0);
  //   // Create an adjacency list.
  //   lchildren.resize(contigs.size());
  //   rchildren.resize(contigs.size());
  //   //int root_contig = -1;
  //   for (int i = 0; i < contigs.size(); ++i) {
  //     if (contigs[i].root == -1) {
  //       //  root_contig = i;
  //       continue;
  //     }
  //     if (contigs[i].dir) {
  //       lchildren[contigs[i].root].push_back(i);
  //     } else {
  //       rchildren[contigs[i].root].push_back(i);
  //     }
  //   }

  //   // Contig enrichment
  //   for (int k = 0; k < 1; ++k) {
  //     total_candidate = 0;
  //     success = 0;
  //     for (int i = contigs.size() - 1; i >= 0; --i) {
  //       //for (int i = 0; i < contigs.size(); ++i) {
  //       EnrichContig(contigs[i].id);
  //     }
  //     cout << (double) success / total_candidate << endl;
  //   }
  //  PrintReadMapping(contigs[3].read_map);
  // PrintReadMapping(contigs[1555].read_map);
  VisitGraph();
  IntTripleVec contig_lens;
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].obsolete) continue;
    
    // CAUTION: somewhere we have a miscalculation. We are always getting one
    // more than the actual lclen.
    //--lclen[i];
    contig_lens.push_back(IntTriple(lclen[i], i, lpath[i].size()));
  }
  sort(contig_lens.begin(), contig_lens.end(), gt_contig_lens);
  cout << "total: " << contig_lens.size() << endl;
  for (int t = 0; t < contig_lens.size(); ++t) {
    int i = contig_lens[t].second;
    if (contigs[i].obsolete) continue;
    cout << ">len: " << lclen[i] << "(" << lpath[i].size() << ")" << endl;
    int l = 0;
    int total_len = 0;
    for (int k = 0; k < lpath[i].size(); ++k) {
      if (contigs[lpath[i][k].first].obsolete) continue;
      contigs[lpath[i][k].first].curr_pos = i;
      total_len += contigs[lpath[i][k].first].size();
      cout << "<" << lpath[i][k].first << "." << lpath[i][k].second << "."
	   << contigs[lpath[i][k].first].size() << ">";
      lpath[i][l++] = lpath[i][k];
      if (lpath[i][k].first != i) contigs[lpath[i][k].first].obsolete = true;
    }
    cout << endl;
    cout << ">total len: " << total_len << endl;
    lpath[i].resize(l);
  }
  cout << "starting read merge\n";
  for (int k = 0; k < contig_lens.size(); ++k) {
    int i = contig_lens[k].second;
    if (contigs[i].obsolete) continue;
    if (lclen[i] < Global::DESIRED_CONTIG_LEN) break;
    IntIntMtTripleVec read_map;
    for (int j = 0; j < lpath[i].size(); ++j) {
      const IntIntMtTripleVec& tr = contigs[lpath[i][j].first].read_map;
      for (int l = 0; l < tr.size(); ++l) {
	//assert(lpath[i][j].second >= 0);
	//assert(tr[l].second >= 0);
	read_map.push_back(IntIntMtTriple(tr[l].first,
					  tr[l].second + lpath[i][j].second,
					  tr[l].third));
      }
    }
    ShortyString c;
    UtilsString::MergeReads(read_map, lclen[i], rp, &c);
    sort(read_map.begin(), read_map.end(), SortReadMap());
    IntIntMtTripleVec::iterator new_end_pos = unique(read_map.begin(),
						     read_map.end(),
						     eq_read_map);
    read_map.erase(new_end_pos, read_map.end());
    //     contigs[i].ext_read_map = read_map;
    //     contigs[i].ext_str = c;
    contigs[i].read_map = read_map;
    contigs[i].str = c;
  }
  fstream* fecontigs = UtilsFile::OpenFile(PREFIX + "contigs-enriched" + ext,
					  "w");
  fstream* fegeom = UtilsFile::OpenFile(PREFIX + "geom-enriched" + ext, "w");
  fstream* feread_info = UtilsFile::OpenFile(PREFIX + "geo-enriched" + ext, "w");
  int total_econtig = 0;
  cout << "starting obsolete detection\n";
  for (int k = 0; k < contig_lens.size(); ++k) {
    int i = contig_lens[k].second;
    //cout << "obsolete detection: " << i << endl;
    if (contigs[i].obsolete) continue;

    // Find more obsolete contigs.
    for (int j = 0; j < lpath[i].size(); ++j) {
      int x = lpath[i][j].first;
      //      if (contigs[x].obsolete) continue;
      for (int l = 0; l < adj[x].size(); ++l) {
	int y = adj[x][l].first;
	if (contigs[y].obsolete) continue;
	//if (lpath[y].size() > 1) continue;
	int offseta, bundle;
	if (OverlapTest(x, y, &offseta, &bundle)) {
	  if (offseta + contigs[y].size() + lpath[i][j].second <=
	      contigs[i].size()) {
	    contigs[y].obsolete = true;
	  } else {
	      IntIntMtTripleVec ar, br;
	      FindCommonReads(contigs, x, y, &ar, &br);
	      if (FractionCovered(y, br) > 98.0) {
	      //if (br.size() == contigs[y].read_map.size()) {
		cout << "hurrahhhhh\n";
		contigs[y].obsolete = true;
	      }
	  }
	}
      }
    }
    //    cout << "obsolete detection: " << i << endl;
    IntVec cut_points;
    ChimeraTest(contigs[i].read_map, i, &cut_points);
    assert (cut_points.size() % 2 == 0);
    cut_points[0] = 0;
    cut_points[cut_points.size() - 1] = lclen[i];
    const ShortyString& str = contigs[i].str.Str();
    int prindex = 0;
    for (int j = 1; j < cut_points.size(); ++j) {
      if (j % 2 == 0) {
	cout << "throwing a segment of size: "
	     << cut_points[j] - cut_points[j - 1] << endl;
	continue; // This is a bad portion.
      }
      while (prindex < contigs[i].read_map.size() &&
             contigs[i].read_map[prindex].second
             < cut_points[j - 1]) {
        ++prindex;
      }
      int rstart = prindex;
      while (prindex < contigs[i].read_map.size() &&
	     contigs[i].read_map[prindex].second +
	     rp->GetMarLen(contigs[i].read_map[prindex].first,
			   contigs[i].read_map[prindex].third)
	     <= cut_points[j]) {
	++prindex;
      }
      int rend = prindex - 1;
      ShortyString pstr = str.substr(cut_points[j - 1],
				     cut_points[j] - cut_points[j - 1]);
      if (pstr.size() < Global::DESIRED_CONTIG_LEN) continue;
      // Replace the trailing dots.
      int l;
      for (l = pstr.size() - 1; l >= 0; --l) {
	if (pstr[l] != '.') break;
      }
      pstr = pstr.substr(0, l + 1);
      if (pstr.size() == 0) continue;
      for (l = 0; l < pstr.size(); ++l) {
 	if (pstr[l] == '.') break;
      }
      if (l < pstr.size()) continue;
      //      *fecontigs << ">contig." << total_econtig << "_" << i << "_" << j - 1
      //		 << endl << pstr << endl; 
      //  *fegeom << ">contig_" << total_econtig << endl << total_econtig << " "
      //	      << pstr << " -1 0 -1" << endl;
      // pstr.PrintScores(*fegeom);
      contig tmp;
      tmp.str = pstr;
      tmp.obsolete = false;
      econtigs.push_back(tmp);
      //      *feread_info << total_econtig << " " << rend - rstart + 1 << endl;
      while (rstart <= rend) {
	assert(contigs[i].read_map[rstart].second >= cut_points[j - 1]);
	econtigs[total_econtig].read_map.push_back(IntIntMtTriple(contigs[i].read_map[rstart].first,
								  contigs[i].read_map[rstart].second - cut_points[j - 1],
								  contigs[i].read_map[rstart].third));
	//*feread_info << contigs[i].read_map[rstart].first << " "
	//	     << contigs[i].read_map[rstart].second - cut_points[j - 1]
	//	     << " " << contigs[i].read_map[rstart].third << " ";
	++rstart;
      }
      sort(econtigs[total_econtig].read_map.begin(),
	   econtigs[total_econtig].read_map.end(), SortReadMap());
      econtigs[total_econtig].read_map_id = econtigs[total_econtig].read_map;
      sort(econtigs[total_econtig].read_map_id.begin(),
	   econtigs[total_econtig].read_map_id.end(), SortReadMapByID());
      //      *feread_info << endl;
      ++total_econtig;
    }
  }
  contigs.resize(econtigs.size());
  contigs = econtigs;
  for (int i = 0; i < contigs.size(); ++i) {
    contigs[i].id = i;
  //contigs[i].root = -1;
  //contigs[i].dist = 0;
  //contigs[i].bundle = 0;
  //contigs[i].obsolete = false;
  //contigs[i].scratch = -1;
  }
  //for (int i = 0; i < contigs.size(); ++i) {
  //for (int j = 0; j < contigs.size(); ++j) {
  //  if (i == j) continue;
  //  int d, sd, bundle;
  //  IntIntMtTripleVec ar, br;
  //  FindCommonSpanningReads(contigs, i, j, &ar, &br);
  //  if (ar.size() < 50) continue;
  //  d = Dist(ar, br, &sd, &bundle);
  //  if (d == INF || bundle < 50) continue;
  //  if (bundle > contigs[j].bundle) {
  //contigs[j].root = i;
  //contigs[j].bundle = bundle;
  //contigs[j].dist = d;
  //contigs[j].sdev = sd;
  //  }
  //  cout << "[" << i << ", " << j << "] ~> [" << bundle << ", " << d << ", " << sd
  //   << "]" << endl;
  //}
  //}
  //for (int i = 0; i < contigs.size(); ++i) {
  //UpdatePath(i);
  // }
  //sort(contigs.begin(), contigs.end(), SortContigs);
  //for (int i = 0; i < contigs.size(); ++i) {
  //for (int j = 0; j < contigs.size(); ++j) {
  //  if (contigs[i].id == contigs[j].id) continue;
  //  if (contigs[i].root == contigs[j].id ||
  //  contigs[i].root == contigs[j].root) {
  //if (contigs[i].dist - contigs[j].dist < contigs[j].size()) {
  //  IntIntMtTripleVec ar, br;
  //  FindCommonReads(contigs, j, i, &ar, &br);
  //  double f = FractionCovered(i, br);
  //  cout << "fraction: " << contigs[j].id << " " << contigs[i].id
  //       << " " << f << endl;
  //  if (f > 98.0) {
  //    cout << "got one\n";
  //    // contigs[i] = contigs[i - 1];
    //    contigs[i].obsolete = true;
  //  }
  //}
  //  }
  //}
  //}
  for (int i = 0; i < contigs.size(); ++i) {
    if (contigs[i].obsolete) continue;
    //  cout << contigs[i].id << " " << contigs[i].root << " " << contigs[i].dist << endl;
    *fecontigs << ">contig." << contigs[i].id << "_" << i << "_" << "."
	       << endl << contigs[i].str << endl; 
    *fegeom << ">contig_" << contigs[i].id << endl << contigs[i].id << " "
	    << contigs[i].str << " -1 0 -1" << endl;
    contigs[i].str.PrintScores(*fegeom);
    *feread_info << contigs[i].id << " " << contigs[i].read_map.size() << endl;
    for (int j = 0; j < contigs[i].read_map.size(); ++j) {
      *feread_info << contigs[i].read_map[j].first << " "
		   << contigs[i].read_map[j].second << " "
		   << contigs[i].read_map[j].third << " ";
    }
    *feread_info << endl;
  }
  vector<contig>().swap(contigs);
  UtilsFile::CloseFile(fecontigs);
  UtilsFile::CloseFile(fegeom);
  UtilsFile::CloseFile(feread_info);
  return 0;
}
