#ifndef SHORTY_STRING_H
#define SHORTY_STRING_H

#include "header.h"
#include "homo-poly.h"
#include "macros.h"
#include "vote.h"

//class HomoPoly;
//class Vote;

// Basically this is a wrapper around the existing string class. This will
// provide the same functionality but at the same time facilitating
// mantaining quality score and voting. One thing to note: this is mainly
// focused on genome alphabet: A, T, G, C.
class ShortyString {
 public:
#ifdef COMPRESS
  typedef vector<HomoPoly> HomoPolyVec;
#endif  // COMPRESS
  typedef vector<Vote> VoteVec;

  ShortyString();
  ShortyString(const string& str);
  ShortyString(const string& str, int x);
  ShortyString(const string& str, const VoteVec& votes);
#ifdef COMPRESS
  ShortyString(const string& str, const HomoPolyVec& hp);
  ShortyString(const string& str, const HomoPolyVec& hp, const VoteVec& votes);
#endif  // COMPRESS
  ~ShortyString();
  inline string Str() const { return str_; }
  inline void Str(const string& str) { str_ = str; }
  inline const VoteVec& Votes() const { return votes_; }
  inline void Votes(const VoteVec& votes) { votes_ = votes; }
  inline const Vote& GetVote(int i) const { return votes_[i]; }
  int GetMaxVote() const;
#ifdef COMPRESS
  inline const HomoPolyVec& HomoPolies() const { return homo_poly_; }
  inline void HomoPolies(const HomoPolyVec& hpv) { homo_poly_ = hpv; }
  inline const HomoPoly& GetHomoPoly(int i) const { return homo_poly_[i]; }
#endif  // COMPRESS
  IntVec RepresentativeVotes() const;

  // Returns a clone but only keeps the bases that we are confident about. We
  // compare the votes of the representative base with that of its neighbors.
  ShortyString GetConfidentClone() const;
  void Print() const;  // This will also print the voting info.
  void PrintScores(ostream& fout) const; // Dump scores so that they are usable.
  void ReadVotesFromFile(istream& fin);

  // Return the reverse complement.
  ShortyString ReverseComplement() const;  

  // Defining method similar to the stl string class.
  void push_back(char c);
  void push_back(char c, const Vote& v);
#ifdef COMPRESS
  void push_back(char c, const HomoPoly& hp, const Vote& v);
#endif  // COMPRESS
  void reverse();
  inline int size() const { return str_.size(); }
  ShortyString substr(int pos, int len) const;
   
  // Operator overloading.
  ShortyString operator + (const ShortyString& ss); 
  ShortyString operator + (const string& s); 
  void operator += (const ShortyString& ss);
  void operator += (const string& s);
  void operator = (const string& str);
  bool operator == (const ShortyString& ss) const;
  bool operator == (const string& str) const;
  bool operator != (const string& str) const;
  char operator [] (int i) const;

  // CAUTION: this is not going to update 'vote_' automatically. Do it manually.
  char& operator [] (int i);
  friend istream& operator >> (istream& is, ShortyString& ss);
  friend ostream& operator << (ostream& os, const ShortyString& ss);

 private:
  string str_;
#ifdef COMPRESS
  HomoPolyVec homo_poly_;
#endif  // COMPRESS
  VoteVec votes_;

  // Other helper methods.
  //   bool EquBaseCount(const BaseCount& a, const BaseCount& b) const; */
  //   bool NEquBaseCount(const BaseCount& a, const BaseCount& b) const; */
  //  void init();
};

// Few typedefs.
typedef vector<ShortyString> SStrVec;
typedef pair<ShortyString, int> SStrIntPair;
typedef vector<SStrIntPair> SStrIntPairVec;
typedef quad<ShortyString, int, int, int> SStrIntIntIntQuad;
typedef vector<SStrIntIntIntQuad> SStrIntIntIntQuadVec;


#endif  // SHORTY_STRING_H
