#ifndef UTILS_STRING_H
#define UTILS_STRING_H

#include "header.h"

class KmarTable;
class ReadProcessor;
class ShortyString;
class Vote;

// CAUTION: make sure that every functin call 'Initialize' first.
class UtilsString {
 public:
  // Align 'b' with 'a' (from offset till len'th base) where 'a' is the
  // reference sequence. So, no modification allow in 'a'. Returns a score.
  // start: starting position of the aligned string in 'a'
  // match: # of matches
  // insertion: # of insertions made to align
  // homo_insertion: how many of the insertiosn were homo-polymer
  // deletion: # of deletions made to align
  // homo_deletion: how many of the deletions were homo-polymer
  // substitution: # of substitutions made to align (for future)
  // report_del: if yes, it prints a '.' for a deletion.
  static int AlignWithReference(bool report_del, int offset, int len,
				const string& a, const string& b,
				int* start, int* match, int* insertion,
				int* homo_insertion, int* deletion,
				int* homo_deletion, string* c);

  // Takes a base 'c' and returns an int: (a, t, g, c)->(0, 1, 2, 3).
  static int Base2Int(char c);

  static base_types Char2Bt(char c);
  
  // Return the compliment of a sequence. Sometimes we need this because we want
  // to manipulate a sequence with the orientation 5'.....3'.
  static string Complement(const string& str);

  // Return the complement of a character.
  static char Complement(char c);

  // This is used to compress a string to remove homo-polymars. Currently used
  // for Helicos data.
  static string CompressStr(const string& str);

  // Counts the number of mismatches between 'a' and 'b'.
  // 'a' and 'b' must have equal length, otherwise returns -1.
  static int CountMisMatch(const string& a, const string& b);
  static int CountMisMatch(const ShortyString& a, const ShortyString& b);

  // Counts the number of mismatches between 'a + offseta' and
  // 'b + offsetb' for 'len' chanracters. 
  static int CountMisMatch(const string& a, const string& b,
			   int offseta, int offsetb, int len);
  static int CountMisMatch(const ShortyString& a, const ShortyString& b,
			   int offseta, int offsetb, int len);

  // Cut the two ends of 'str' and return the pruned string. Also store the two
  // ends in 'left' and 'right'. Note that 'left' and 'right' will have some
  // overlap the new string.
  static ShortyString CutTails(const ShortyString& str, ShortyString* left,
			 ShortyString* right);

  // Check whether 'a' and 'b' are eligible for overlap. It returns true if 'a'
  // 'b' has a common mar of size 'MIN_CONSEQ_OVERLAP'. This will take linear
  // time whereas 'Overlap' takes quadratic time. Thats how we are going to save
  // time. Also we are using the positional/generation information.
  static bool Eligible4Overlap(int amin, int amax, int aming, int amaxg,
			       int bmin, int bmax, int bming, int bmaxg,
			       int gen_dev, const string& a, const string& b);
  static bool Eligible4Overlap(int amin, int amax, int aming, int amaxg,
			       int bmin, int bmax, int bming, int bmaxg,
			       int gen_dev, const ShortyString& a,
			       const ShortyString& b);

  // Check Eligibility based on a pre-matched kmar.
  static bool Eligible4Overlap(int amin, int amax, int aming, int amaxg,
			       int bmin, int bmax, int bming, int bmaxg,
			       int offseta, int offsetb, int k, const string& a,
			       const string& b);

  // Takes a fasta style string and converts it to AB's Solid-style data. But in
  // order to use it inside our assembler we represent {0, 1, 2, 3} by {A, T, G,
  // C}
  static string Fasta2Solid(const string& a);

  // Finds the best position for alignment in 'a' for 'b'.
  static int FindBestPos(const string& a, const string& b,
			 const KmarTable& ktable, int* score, int* match,
			 int* ins, int* hins, int* del, int* hdel, string* c);

  // Trying to align 'b' against 'a' or the reverse complement of 'a'.
  static void GetBestAlign(const string& a, const string& b,
			   int offset, int len, bool* rc, int* score,
			   int* start, int* match, int* ins, int* hins,
			   int* del, int* hdel, string* c);

  // Same as before but uses a kmar-table, so no need of offset and len.
  static void GetBestAlign(const string& a, const string& b, const KmarTable& ktable,
		    const KmarTable& ktablerc, bool* rc, int* score, int* start,
		    int* match, int* ins, int* hins, int* del, int* hdel,
		    string* c);

  // Used to initialize 'base2int_' & 'int2base_'.
  static void Initialize();

  // Takes an int 'x' and returns the corresponding base: (0, 1, 2, 3)->
  // (A, T, G, C).
  static char Int2Base(int x);

  // Take an integer and converts that to a string, nothing to do with genome
  // alphabet.
  static string Int2Str(int x);

  // Merge two contigs 'a' and 'b'. Please note that 'a' will remain in the left
  // side of the merge contig. This is used where we don't want additional
  // information, only see whether the two string can be merged or not.
  static ShortyString OnlyMergeWithoutDel(const ShortyString& a,
					  const ShortyString& b);

  // Same as before, but puts 'b' at offset 'offseta' of 'a'. So much simpler. 
  static ShortyString OnlyMergeWithoutDel(const ShortyString& a,
					  const ShortyString& b, int offseta);

  static int OverlapWithoutDel(const ShortyString& a, const ShortyString& b,
			       int* offseta, int* score, ShortyString* c );

  // Generate a string by merging reads mentioned in 'reads'. It has information
  // in the form [read id, offset, mar type]. It fetches the reads from 'rp'.
  static void MergeReads(const IntIntMtTripleVec& reads,
			 int elen, const ReadProcessor* rp,
			 ShortyString* c);

  // Tries to merge two contigs represented by reads in 'a' and 'b'. If
  // successfull the string is in 'c' and it returns true. Otherwise false is
  // returned. Reads in 'a' and 'b' are expected to be sorted by
  // offset->id->mar_type. It will only try to merge a-b, not b-a.
  static bool OverlapFromReadMapping (const ShortyString& a,
				      const ShortyString& b,
				      const IntIntMtTripleVec& ar,
				      const IntIntMtTripleVec& br,
				      const ReadProcessor* rp,
				      int* offseta, int* bundle,
				      ShortyString* c,
				      IntIntMtTripleVec* cr);

  static bool OverlapFromCommonReadMapping(const ShortyString& a,
					   const ShortyString& b,
					   const IntIntMtTripleVec& ar,
					   const IntIntMtTripleVec& br,
					   const ReadProcessor* rp,
					   int* offseta, int* bundle,
					   ShortyString* c);

  // Calculate the overlap between 'a' and 'b'. 'score' is the number of
  // mathces within the overlap. We dont count 'N' or '?' as actual match.
  // 'c' contains the merged string from 'a' and 'b'.
  static int Overlap(const string& a, const string& b, int* score, string* c,
		     bool allow_deletion);

  // Exactly same as the previous one but excludes Harvard specific issues like
  // '?'. This is mainly intended for 'Helicos' and 'AB' data. In essence, this
  // should be simpler than the other 'Overlap'. 'offseta' is the number of
  // bases at the beginning of 'a' which is not overlapped by 'b'. This info is
  // useful for geography detection.
  static int Overlap(const ShortyString& a, const ShortyString& b, int* offseta,
		     int* score, ShortyString* c);

  // Print contig in fasta format.
  static void PrintContig(const string& id, const string& str);
  static void PrintContig(const string& id, const ShortyString& str);

  // Tries to map a 'read' in a seed with some error. Strict matching fails for
  // dataset like Helicos.
  static bool ReadMapsInSeed(int offset, int len, const ShortyString& a,
			     const ShortyString& b);

  // Return the reverse of a sequence.
  static ShortyString Reverse(const ShortyString& str);
  static string Reverse(const string& str);

  // Return the reverse complement of a sequence.
  static ShortyString ReverseComplement(const ShortyString& str);
  static string ReverseComplement(const string& str);

  // This is used to calculate the hash value of a right shifted, previously
  // hashed string, 'gone' is the first character in the previous string, 'come'
  // is the new character at the end, 'prev' is the old hash value. 'window' is
  // the windows size that is used to calculate the value (for hashing: hash
  // length).
  static int RightShiftStrVal(char gone, char come, int prev, int window);

  // Takes a AB style color-space input and returns what we understand :)
  // e.g. T0123012013
  static string Solid2Fasta(const string& a);

  // Convert 'size' # of characted starting from 'offset' to an int.
  static int String2Int(const string& a, int offset, int size);

  // Checks whether 'p' is a substring of 't'. Uses KMP algorithm.
  static bool SubStr(const string& t, const string& p);

 private:
  static bool initialized_;
  //  static bool deleted1_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP];
  static char int2base_[BASE_TYPES];
  static char charchar2color_[MAX_CHAR][MAX_CHAR]; // CAUTION: char type data.
  static char charcolor2char_[MAX_CHAR][4];
  static char comp_[MAX_CHAR];
  static char delcost_[MAX_CHAR];  // CAUTION: char type data.
  static char subscost_[MAX_CHAR][MAX_CHAR];  // CAUTION: char type data.
  static char subschar_[MAX_CHAR][MAX_CHAR];
  static int min_overlap_score_;
  static int base2int_[MAX_CHAR];
  //  static int deleteda_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP];
  //static int deletedb_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP];
  //static int matrix_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP];
  //static int pi_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP][2];
  //  static int scratch_[1 << (MIN_CONSEQ_OVERLAP << 1)];
  static int similarity_[MAX_CHAR][MAX_CHAR];
  //static int subs_[MAX_LEN4OVERLAP][MAX_LEN4OVERLAP];
  static base_types char2bt_[MAX_CHAR];
  static vector<Vote> scratch_votes;

/*   // Update 'matrix_' in 'Overlap'. */
/*   static void UpdateMatrix(int i, int j, int i_, int j_, int t, int dela, */
/* 			   int delb) { */
/*     if (t >= matrix_[i][j]) { */
/*       matrix_[i][j] = t; */
/*       deleteda_[i][j] = dela; */
/*       deletedb_[i][j] = delb; */
/*       pi_[i][j][0] = i_; */
/*       pi_[i][j][1] = j_; */
/*     } */
/*   } */
// Helper method for OverlapFromCommonReadMapping.
  struct SortCommonReads {
    bool operator() (const IntIntMtTriple& a, const IntIntMtTriple& b) {
      if (a.second == b.second) return a.first < b.first;
      return a.second < b.second;
    }
  };
};

#endif  //  UTILS_STRING_H
